
from __future__ import print_function

import argparse
import os, datetime
import numpy as np

from tensorboardX import SummaryWriter
import torch
from torch.utils.data import TensorDataset, DataLoader
from torch.utils.data.distributed import DistributedSampler
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

curr_path = os.path.dirname(os.path.abspath(__file__))

### Environment variable initialization
# The machine with rank 0 will be used to set up all connections.
RANK = int(os.environ.get('RANK', 0))                       # required; can be set either here, or in a call to init function
WORLD_SIZE = int(os.environ.get('WORLD_SIZE', 1))           # required; can be set either here, or in a call to init function
MASTER_ADDR = os.environ.get('MASTER_ADDR', 'localhost')    # required (except for rank 0); address of rank 0 node
MASTER_PORT = os.environ.get('MASTER_PORT', "23456")        # required; has to be a free port on machine with rank 0

def should_distribute():
    return dist.is_available() and WORLD_SIZE > 1

def is_distributed():
    return dist.is_available() and dist.is_initialized()

### 1. Data Loading
def load_data(train_batch_size, test_batch_size):
    '''
    Load X_train.npy y_train.npy X_val.npy y_val.npy
    '''
    X_train = torch.from_numpy(np.load(os.path.join(curr_path, './X_train.npy')))
    y_train = torch.from_numpy(np.load(os.path.join(curr_path,'./y_train.npy')))
    X_val = torch.from_numpy(np.load(os.path.join(curr_path,'./X_val.npy')))
    y_val = torch.from_numpy(np.load(os.path.join(curr_path,'./y_val.npy')))
    
    train_dataset = TensorDataset(X_train, y_train)

    # use torch.utils.data.DataLoader.distributed.DistributedSampler for distributed training
    # so that each training node has different sample of data
    train_sampler = DistributedSampler(train_dataset, num_replicas=WORLD_SIZE, rank=RANK) if is_distributed() else None

    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=train_batch_size,
        shuffle=(train_sampler is None),
        sampler=train_sampler,
        )

    test_dataset = TensorDataset(X_val, y_val)
    test_loader = DataLoader(
        dataset=test_dataset,
        batch_size=test_batch_size,
        shuffle=False,
        )
    return train_loader, test_loader, train_sampler

### 2. Define your network architecture
class Net(nn.Module):
    def __init__(self, input_dimension):
        super(Net, self).__init__()
        '''
        define layers
        '''
        self.linear = nn.Linear(input_dimension, 2)

    def forward(self, x):
        '''
        use layers to process input x
        '''
        x = self.linear(x)
        return x

def train(model, device, train_loader, loss_function, epoch, writer, optimizer):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        # set gradients to zero
        optimizer.zero_grad()
        output = model(data)
        loss = loss_function(output, target)
        # compute the gradients
        loss.backward()
        # updates the parameters
        optimizer.step()

        print('Train Epoch: {} [{}/{} ({:.0f}%)]\tloss={:.4f}'.format(
            epoch, batch_idx * len(data), len(train_loader.dataset),
            100. * batch_idx / len(train_loader), loss.item()))
        if RANK == 0:
            writer.add_scalar('loss', loss.item(), epoch * len(train_loader) + batch_idx)


def test(model, device, test_loader, loss_function, epoch, writer):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += loss_function(output, target).item() # sum up batch loss
            pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    print('\naccuracy={:.4f}\n'.format(float(correct) / len(test_loader.dataset)))
    if RANK == 0:
        writer.add_scalar('accuracy', float(correct) / len(test_loader.dataset), epoch)

def main():
    torch.manual_seed(1)

    parser = argparse.ArgumentParser(description='PyTorch Distributed Training Example')
    parser.add_argument('--batch-size', type=int, default=128, metavar='N',
                        help='input batch size for training (default: 128)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='TN',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=20, metavar='EP',
                        help='number of epochs to train (default: 20)')
    parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--dir', default=os.path.join(curr_path, 'output', 'runs', datetime.datetime.now().strftime("%Y%m%d_%H%M%S")), 
                        metavar='L', 
                        help='directory where output model are stored')

    # 0. cpu/gpu settings, distributed settings
    if dist.is_available():
        parser.add_argument('--backend', type=str, help='Distributed backend',
                            choices=[dist.Backend.GLOO, dist.Backend.NCCL, dist.Backend.MPI],
                            default='auto')
    args = parser.parse_args()

    os.system("nvidia-smi")
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")

    # Set backend according to rule of thumb if it's not specified by user
    # Rule of thumb: NCCL for GPU, GLOO for CPU
    if args.backend == 'auto':
        if device == torch.device("cuda"):
            args.backend = dist.Backend.NCCL
        else:
            args.backend = dist.Backend.GLOO
    
    # Check the used device
    print('Using with {}'.format(device))

    '''
    use dist.init_process_group to initialize the process group. Wait until all processes have joined.
    '''
    if should_distribute():
        print('Using distributed PyTorch with {} backend'.format(args.backend))
        # Initializes the default distributed process group, and this will also initialize the distributed package.
        dist.init_process_group(
            backend=args.backend, 
            init_method="env://",
            world_size=WORLD_SIZE,
            rank=RANK,
            )

    # 1. load data
    train_loader, test_loader, train_sampler = load_data(args.batch_size, args.test_batch_size)

    # 2. load model
    origin_model = Net(input_dimension=5).to(device)

    if is_distributed():
        '''
        TODO use DistributedDataParallel to parallelizes the model by splitting the input across the specified devices by chunking in the batch dimension. 
        The module is replicated on each machine and each device, and each such replica handles a portion of the input. 
        During the backwards pass, gradients from each node are averaged.
        '''
        Distributor = nn.parallel.DistributedDataParallel if use_cuda \
            else nn.parallel.DistributedDataParallelCPU
        model = Distributor(origin_model)
    else:
        model = origin_model

    # 3 TensorBoard writer / optimizer / loss_function
    print("Tensorboard save logs to", args.dir)
    writer = SummaryWriter(args.dir)
    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)
    loss_function = nn.CrossEntropyLoss(
        weight=torch.tensor([1, 2], dtype=torch.float, device=device) # for unbalanced dataset, you can change weight
    )

    # 4. Training and testing
    for epoch in range(1, args.epochs + 1):
        # DistributedSampler is not used in official MNIST examples: https://github.com/kubeflow/training-operator/tree/master/examples/pytorch/mnist
        # which will cause all the distributed nodes using the same data with the same model (won't be faster) 
        # DistributedSampler doc: https://pytorch.org/docs/stable/data.html
        if is_distributed():
            train_sampler.set_epoch(epoch)
        train(model, device, train_loader, loss_function, epoch, writer, optimizer)
        test(model, device, test_loader, loss_function, epoch, writer)
    
    os.system("nvidia-smi")
    if RANK == 0:
        print("saving model to", args.dir)
        os.makedirs(args.dir, exist_ok=True) 
        torch.save(model.state_dict(), os.path.join(args.dir, "pytorch_one_layer.pt"))


    # A little testing
    print()
    print("Inspect the model")
    print("The parameters of the model:")
    print(list(origin_model.linear.named_parameters()))
    
    print()
    print("Try predict some data:")
    # recall the features are: [short_text, body, business, html, money]
    print("Predict result: 0: ham, 1: spam")
    for case in [    
            [0, 0, 0, 0, 0], 
            [1, 0, 0, 0, 0], 
            [1, 0, 1, 1, 1], 
            [0, 1, 0, 0, 1], 
            [0, 1, 1, 1, 1],
            [1, 1, 1, 1, 1]]:
        case_tensor = torch.tensor(case, dtype=torch.float32, device=device).reshape(1, -1)
        pred = model(case_tensor).max(1, keepdim=True)[1]
        print("-"*10)
        print("The test case:", case)
        print("The predict result", pred.item())

if __name__ == '__main__':
    main()
                
